@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1999, 2017, Oracle and/or its affiliates. All rights reserved.
+ * Copyright (c) 1999, 2019, Oracle and/or its affiliates. All rights reserved.
  * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
  *
  * This code is free software; you can redistribute it and/or modify it
@@ -36,19 +36,21 @@ class VersionProps {
         "openjdk";
 
     private static final String java_version =
-        "11.0.7";
+        "11.0.8";
 
     private static final String java_version_date =
-        "2020-04-14";
+        "2020-07-14";
 
+    // This field is read by HotSpot
     private static final String java_runtime_name =
         "OpenJDK Runtime Environment";
 
+    // This field is read by HotSpot
     private static final String java_runtime_version =
-        "11.0.7+10";
+        "11.0.8+10";
 
     private static final String VERSION_NUMBER =
-        "11.0.7";
+        "11.0.8";
 
     private static final String VERSION_BUILD =
         "10";
@@ -62,12 +64,25 @@ class VersionProps {
     private static final boolean isLTS =
         "".startsWith("LTS");
 
-    private static final String VENDOR_VERSION_STRING =
+    private static final String VENDOR =
+        "N/A";
+
+    private static final String VENDOR_URL =
+        "https://openjdk.java.net/";
+
+    // The remaining VENDOR_* fields must not be final,
+    // so that they can be redefined by jlink plugins
+
+    // This field is read by HotSpot
+    private static String VENDOR_VERSION =
         "";
 
-    private static final String vendor_version =
-        (!VENDOR_VERSION_STRING.isEmpty()
-         ? " " + VENDOR_VERSION_STRING : "");
+    private static String VENDOR_URL_BUG =
+        "https://bugreport.java.com/bugreport/";
+
+    // This field is read by HotSpot
+    private static String VENDOR_URL_VM_BUG =
+        "https://bugreport.java.com/bugreport/crash.jsp";
 
     static {
         init();
@@ -78,8 +93,12 @@ class VersionProps {
         System.setProperty("java.version.date", java_version_date);
         System.setProperty("java.runtime.version", java_runtime_version);
         System.setProperty("java.runtime.name", java_runtime_name);
-        if (!VENDOR_VERSION_STRING.isEmpty())
-            System.setProperty("java.vendor.version", VENDOR_VERSION_STRING);
+        if (!VENDOR_VERSION.isEmpty())
+            System.setProperty("java.vendor.version", VENDOR_VERSION);
+
+        System.setProperty("java.vendor", VENDOR);
+        System.setProperty("java.vendor.url", VENDOR_URL);
+        System.setProperty("java.vendor.url.bug", VENDOR_URL_BUG);
     }
 
     private static int parseVersionNumber(String version, int prevIndex, int index) {
@@ -189,6 +208,9 @@ class VersionProps {
             jdk_debug_level = jdk_debug_level + " ";
         }
 
+        String vendor_version = (VENDOR_VERSION.isEmpty()
+                                 ? "" : " " + VENDOR_VERSION);
+
         ps.println(java_runtime_name + vendor_version
                    + " (" + jdk_debug_level + "build " + java_runtime_version + ")");
 
